"""Utility classes and functions to deal with the byoa manifest file"""

from functools import wraps
from pathlib import Path
from typing import Any, Optional

import toml

from byoa import MANIFEST_FILE
from byoa.exceptions import NoManifestFoundException, UnknownKeyException


def check_manifest(function):
    """Checks wether the byoa manifest file exists in the current directory

    Use this decorator to ensure that no statement is executed if the manifest file is missing.
    """

    @wraps(function)  # ensures the docstring is also wrapped
    def wrapper(*args, **kwargs):
        if not Path(MANIFEST_FILE).exists():
            raise NoManifestFoundException

        function(*args, **kwargs)

    return wrapper


class Manifest:
    """Object representation of the byoa manifest file

    Args:
      from_file (bool): if True, will read and load the byoa manifest file.
        Defaults to True.
    """

    POSSIBLE_KEYS = ["name", "slug", "summary", "description", "author"]
    """List of used keys."""

    def __init__(self, from_file=True) -> None:
        if from_file and not Path(MANIFEST_FILE).exists():
            raise NoManifestFoundException

        if not from_file:
            self.__manifest = {}
        else:
            self.__manifest = toml.load(MANIFEST_FILE)

    def set(self, key: str, value):
        """Sets the manifest key/value pair

        Args:
          key (str): The key
          value: The value

        Raises:
          UnknownKeyException: If the key is unknown
        """

        if key not in self.POSSIBLE_KEYS:
            raise UnknownKeyException(key)

        self.__manifest.update({key: value})

    def get(self, key: str) -> Optional[Any]:
        """Returns the value of attribute 'key'

        Args:
          key (str): The key

        Returns:
          Optional[Any]: The value if it exists
        """

        return self.__manifest.get(key)

    def save(self, directory=Path(".")):
        """Saves the manifest as a file

        Args:
          directory (pathlib.Path): The directory in which the file will be saved,
            defaults to the current directory
        """
        with open(directory.joinpath(MANIFEST_FILE), mode="w", encoding="utf-8") as file:
            toml.dump(self.__manifest, file)

    @property
    def name(self):
        """Name of the processor"""
        return self.get("name")

    @property
    def slug(self):
        """Human-readable, unique identifier of the processor"""
        return self.get("slug")

    @property
    def summary(self):
        """A short sentence describing the processor"""
        return self.get("summary")

    @property
    def description(self):
        """Description of the project"""
        return self.get("description")

    @property
    def author(self):
        """Author(s) of the project"""
        return self.get("author")



"""Command line interface of the BYOA tool"""

import click

from byoa import __version__
from byoa.cli.build import build
from byoa.cli.context import context
from byoa.cli.delete import delete
from byoa.cli.deploy import deploy
from byoa.cli.init import init_processor
from byoa.cli.manifest import update_manifest
from byoa.cli.run import run_processor


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(
    __version__, "--version", "-V", package_name="byoa", message="%(package)s %(version)s"
)
def main():
    """A command-line utility to Build Your Own Analytic"""


main.add_command(build)
main.add_command(context)
main.add_command(delete)
main.add_command(deploy)
main.add_command(init_processor)
main.add_command(update_manifest)
main.add_command(run_processor)



"""Input and output base objects"""


import json
from typing import Optional, Self

from pydantic import BaseModel, ValidationError


class InputModel(BaseModel):
    """A base class for defining the input schema of a processor"""

    @classmethod
    def from_file(cls, input_data_path: str) -> Optional[Self]:
        """Load and return the input data from JSON file.

        Args:
          input_data_path(str): Path to the JSON file where the input is stored

        Returns:
          Optional[Self]: The object loaded from the input file if loading was sucessful
        """

        with open(input_data_path, "r", encoding="utf-8") as file:
            input_data = json.load(file)

            try:
                input_model = cls(**input_data)
                return input_model

            except ValidationError as e:
                print(f"Input validation error: {e.errors()}")
                return None


class OutputModel(BaseModel):
    """A base class for defining the ouput schema of a processor"""



"""Logging trace formater """

import logging

from opentelemetry import trace


class TraceFormatter(logging.Formatter):
    """
    Formatter class to include trace information in log records.

    Methods:
        format(record): Formats the log record with trace information.
    """

    def format(self, record):
        """
        Formats the log record with trace information.

        Args:
            record (logging.LogRecord): Log record to be formatted.

        Returns:
            str: Formatted log message including trace information.
        """
        current_span = trace.get_current_span()

        trace_id = current_span.get_span_context().trace_id if current_span else None
        record.trace_id = hex(trace_id) if trace_id else ""

        return super().format(record)



"""AWS S3 helper functions"""

import os

import boto3
from boto3.exceptions import Boto3Error


def upload_file(local_file_path: str, bucket_name: str) -> bool:
    """Upload a file to AWS S3 Bucket.

    Args:
      local_file_path(str): The local file path to upload.
      bucket_name(str): The name of the AWS S3 bucket

    Returns:
      bool: True if upload was sucessful, False otherwise

    """

    if bucket_name is None:
        raise ValueError("bucket_name cannot be 'None'")

    s3_client = boto3.client("s3")
    if s3_client and bucket_name is not None:
        try:
            file_name = os.path.basename(local_file_path)
            s3_client.upload_file(local_file_path, bucket_name, file_name)
            return True
        except Boto3Error as error:
            print(f"Error while uploading file to AWS S3: {error}")
            return False
    else:
        return False


def write_folder_to_aws_s3(local_folder_path: str, bucket_name: str) -> bool:
    """Upload a folder to AWS S3 Bucket.

    Args:
      local_folder_path(str): The local folder path to upload
      bucket_name(str): The name of the AWS S3 bucket

    Returns:
      bool: True if upload was sucessful, False otherwise
    """

    if bucket_name is None:
        raise ValueError("bucket_name cannot be 'None'")

    s3_client = boto3.client("s3")
    if s3_client:
        try:
            for root, _, files in os.walk(local_folder_path):
                for file in files:
                    local_file_path = os.path.join(root, file)
                    relative_path = os.path.relpath(local_file_path, local_folder_path)
                    s3_key = os.path.join(
                        os.path.basename(local_folder_path), relative_path
                    ).replace(os.sep, "/")

                    if bucket_name is not None:
                        s3_client.upload_file(local_file_path, bucket_name, s3_key)
            return True
        except Boto3Error as error:
            print(f"Error while uploading folder to AWS S3: {error}")
            return False
    else:
        return False


def get_s3_uri_path(local_path: str, bucket_name: str) -> str:
    """Get the s3 path of the uploaded element (file or folder)

    Args:
      local_path(str): The local path of the uploaded folder/file on s3
      bucket_name(str): The bucket name set to store the file on s3

    Returns:
      str: the s3 uri of the uploaded folder/file
    """
    # get bucket name
    if bucket_name is None:
        raise ValueError("bucket_name cannot be 'None'")

    s3_key = os.path.basename(local_path)
    s3_uri = f"s3://{bucket_name}/{s3_key}"

    return s3_uri



"""Run command"""

import shlex
import subprocess

import click
import docker

from byoa.config.manifest import Manifest, check_manifest


@click.command("run", short_help="executes the processor")
@click.option(
    "--entrypoint",
    "-e",
    type=click.Choice(["api", "cli"], case_sensitive=False),
    default="cli",
    help="How the processor is started",
)
@click.option(
    "--port",
    "-p",
    default=8081,
    help="Redirection port when using api mode",
)
@click.option("--build", "-b", is_flag=True, help="Build the image before running")
@click.option(
    "--args",
    "-a",
    help="Additional arguments to pass to the command",
)
@check_manifest
def run_processor(entrypoint: str, build: bool, args: str, port: int):
    """Runs the processor's Docker image"""

    manifest = Manifest()
    docker_client = docker.from_env()

    if build:
        click.echo("Building image...", nl=False)
        image, _ = docker_client.images.build(path="./", tag=manifest.slug)
        click.echo("DONE !")
    else:
        image = docker_client.images.get(manifest.slug)

    click.echo(f"Running image {image.tag}")
    if entrypoint == "api":
        docker_client.containers.run(
            image, environment=["RUN_MODE_ENV=API"], ports={"80": port}, detach=True
        )
        click.echo(
            f"API running on http://localhost:{port}. "
            f"Swagger UI available at http://localhost:{port}/docs",
        )

    else:
        if args:
            arg_list = shlex.split(args)
        else:
            arg_list = []

        command = ["docker", "run", image.id] + arg_list

        raw_output = subprocess.run(command, capture_output=True, text=True, check=True).stdout
        click.echo(raw_output)



"""Utility functions to deal with the BYOA context variables"""

import configparser
from pathlib import Path
from typing import Optional

from byoa.config import DEFAULT_CONFIG_DIR

CONTEXT_FILE = "context"
REGISTRY_KEY = "registry"
ENDPOINT_KEY = "endpoint"


def get_context_value(
    key: str,
    profile="default",
    fallback: Optional[str] = None,
    config_dir: Path = DEFAULT_CONFIG_DIR,
) -> Optional[str]:
    """Get value from context file

    Args:
        key (str): Key
        profile (str, optional): Profile name. Defaults to "default".
        fallback (Optional[str], optional): Fallback value if the context value is None.
        Defaults to None.
        config_dir (Path, optional): Path to the configuration directory. Defaults to ~/.byoa

    Returns:
        str: context value
    """

    parser = configparser.ConfigParser()
    parser.read(config_dir.joinpath(CONTEXT_FILE))
    return parser.get(profile, key, fallback=fallback)


def append_context(
    registry: str, endpoint: str, profile: str, config_dir: Path = DEFAULT_CONFIG_DIR
):
    """Adds the context profile to the context file

    Args:
        registry (str): Registry value
        endpoint (str): Endpoint value
        profile (str): Profile
        config_dir (Path, optional): Path to the configuration directory. Defaults to ~/.byoa
    """

    if not config_dir.exists():
        config_dir.mkdir()

    context_file = config_dir.joinpath(CONTEXT_FILE)
    parser = configparser.ConfigParser()
    parser.read(context_file)

    if profile not in parser.sections():
        parser.add_section(profile)

    parser.set(profile, REGISTRY_KEY, registry)
    parser.set(profile, ENDPOINT_KEY, endpoint)

    with open(context_file, "w", encoding="utf-8") as f:
        parser.write(f)



"""Delete command"""

import click
import requests
from click_option_group import RequiredMutuallyExclusiveOptionGroup, optgroup

from byoa.config.context import ENDPOINT_KEY, get_context_value
from byoa.config.manifest import Manifest
from byoa.exceptions import MissingContextValue


@click.command("delete", short_help="deletes the processor deployment")
@click.option(
    "--endpoint",
    help="BYOA API endpoint",
    envvar="BYOA_ENDPOINT",
    show_envvar=True,
    default=lambda: get_context_value(ENDPOINT_KEY),
)
@optgroup.group("Version(s)", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--all", "_all", is_flag=True, help="All versions of the processor")
@optgroup.option("--version", help="Specific version")
def delete(profile: str, endpoint: str, version: str, _all: bool):
    """Dergisters the processor version

    If all versions of the processor are deregistered, the processor will be deregistered.

    Note that the version docker images will not be deleted from their registry.
    """
    if profile:
        _endpoint = get_context_value(ENDPOINT_KEY, profile, endpoint)

        if _endpoint is None:
            raise MissingContextValue(ENDPOINT_KEY)

        endpoint = _endpoint

    manifest = Manifest()

    if _all:
        click.echo("Dergister versions...", nl=False)
        response = requests.delete(
            f"{endpoint}/processors/{manifest.slug}/versions",
            timeout=60,
        )
        click.echo(response.reason)

        click.echo("Dergister processor...", nl=False)
        response = requests.delete(
            f"{endpoint}/processors/{manifest.slug}",
            timeout=60,
        )
        click.echo(response.reason)

    else:
        click.echo(f"Dergister version {version}...", nl=False)
        response = requests.delete(
            f"{endpoint}/processors/{manifest.slug}/versions/{version}",
            timeout=60,
        )
        click.echo(response.reason)



"""Manifest command"""

import click

from byoa.config.manifest import Manifest, check_manifest


@click.command("manifest", short_help="sets manifest key/value pair")
@click.argument("key", type=click.Choice(Manifest.POSSIBLE_KEYS, case_sensitive=False))
@click.argument("value")
@check_manifest
def update_manifest(key: str, value: str):
    """Sets manifest KEY/VALUE pair"""

    manifest = Manifest()
    manifest.set(key, value)
    manifest.save()



"""Build command"""

import os
import subprocess

import click
import docker
import toml

from byoa.config.manifest import Manifest, check_manifest


@click.group("build", short_help="builds the processor")
def build():
    """Build the processor"""


@build.command("image", short_help="builds a processor image")
@check_manifest
def build_image():
    """Builds the processor as a Docker image"""

    manifest = Manifest()
    docker_client = docker.from_env()

    click.echo("Building image...", nl=False)
    docker_client.images.build(path="./", tag=manifest.slug)
    click.echo("DONE !")


@build.command("wheel", short_help="builds a processor wheel")
@check_manifest
def build_wheel():
    """Builds the processor package as a wheel"""

    command = "python -m pip wheel --wheel-dir=dist --no-deps ."
    result = subprocess.run(command, shell=True, capture_output=True, text=True, check=True)

    with open("pyproject.toml", "r", encoding="utf-8") as f:
        config = toml.load(f)

    if result.returncode == 0:
        # output_lines = result.stdout.split()
        click.echo(
            "Setup file generated successfully.\nPlease use the command below to install it:"
        )
        file_install = (
            config["project"]["name"] + "-" + config["project"]["version"] + "-py3-none-any.whl"
        )
        folder_path = "dist/"

        file_path = os.path.join(folder_path, file_install)

        if os.path.exists(file_path):
            click.secho("pip install dist/" + file_install, bold=True)
        else:
            click.secho("pip install dist/your_file.whl", bold=True)

    else:
        click.echo("An error occured during the setup file generation, please see error below:")
        click.echo(result.stderr)
